# coding: utf-8
import datetime

from django.core.exceptions import NON_FIELD_ERRORS
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.query_utils import Q

from educommon.utils.date import date_range_to_str


class DateIntervalMeta(models.base.ModelBase):

    u"""Метакласс для примеси DateIntervalMixin.

    Добавляет к создаваемой модели поля, содержащие границы интервала. Имена
    полей задаются в атрибуте interval_field_names.
    """

    def __new__(cls, name, bases, attrs):
        if 'DateIntervalMixin' in globals() and DateIntervalMixin in bases:
            # сейчас в cls лежит класс, к которому "подмешивается" примесь
            if 'interval_field_names' in attrs:
                # заданы новые имена полей для хранения границ интервала
                interval_field_names = attrs['interval_field_names']
            else:
                interval_field_names = DateIntervalMixin.interval_field_names
            verbose_names = (u'Начало интервала', u'Конец интервала')

            # Добавление полей, хранящих границы интервала
            for field_name, verbose_name in zip(interval_field_names,
                                                verbose_names):
                field = models.DateField(
                    name=field_name,
                    blank=True, null=True,
                    verbose_name=verbose_name,
                )
                attrs[field_name] = field

        return super(DateIntervalMeta, cls).__new__(cls, name, bases, attrs)


class DateIntervalMixin(models.Model):

    u"""Примесь к моделям, добавляющая интервал дат.

    Содержит два поля: date_from (дата начала интервала) и date_to (дата
    окончания интервала). Значение None этих полей указывает на то, что
    временной интервал не ограничен с соответствующей стороны.
    """

    __metaclass__ = DateIntervalMeta

    no_intersections_for = None
    u"""Список имен полей, по которым осуществляется поиск пересечений.

    Если в модели есть другие записи с такой же комбинацией значений указанных
    полей и пересекающимся интервалом, то экземпляр данной модели не пройдет
    валидацию данных.

    Если значением атрибута является None, проверка на пересечение не
    выполняется. Пустой список указывает на необходимость поиска пересечений
    со всеми записями модели.
    """

    interval_field_names = ('date_from', 'date_to')
    u"""Имена полей, хранящих границы интервала.

    При необходимости переименования полей нужно переопределить данный атрибут
    в потомке.

    .. code::
       class TestModel(DateIntervalMixin, BaseModel):
           interval_field_names = ('start', 'end')
    """

    @property
    def interval_range(self):
        u"""Возвращает дану начала интервала и дату окончания интервала.

        :rtype: tuple
        """
        date_from_name, date_to_name = self.interval_field_names

        date_from = getattr(self, date_from_name)
        date_to = getattr(self, date_to_name)

        result = (date_from, date_to)

        return result

    def is_date_in_interval(self, day):
        u"""Возвращает True, если указанная дата входит в интервал.

        :type day: datetime.date

        :rtype: bool
        """
        assert isinstance(day, datetime.date), type(day)

        date_from, date_to = self.interval_range
        result = (
            date_from is None and date_to is None or
            date_from is None and date_to is not None and day <= date_to or
            date_from is not None and date_to is None and day >= date_from or
            date_from is not None and date_to is not None and
            date_from <= day <= date_to
        )
        return result

    @property
    def interval_range_str(self):
        date_from, date_to = self.interval_range
        result = date_range_to_str(date_from, date_to)
        return result

    def get_intersected_query(self):
        u"""Запрос на выборку записей с пересекающимися интервалами.

        :rtype: django.db.models.query.QuerySet
        """
        # pylint: disable=star-args
        if self.no_intersections_for is None:
            return self.__class__.objects.none()

        query = self.__class__.objects.all()

        if self.pk:
            query = query.exclude(pk=self.pk)

        if len(self.no_intersections_for) > 0:
            query = query.filter(**{
                field_name: getattr(self, field_name)
                for field_name in self.no_intersections_for
            })

        # Для проверки пересечения диапазонов используется тип данных
        # PostgreSQL daterange
        cond = (
            "not isempty("
            "daterange(%s, %s, '[]') * daterange({}, {}, '[]')"
            ")"
        ).format(*self.interval_field_names)
        query = query.extra(where=[cond], params=self.interval_range)

        return query

    @classmethod
    def get_date_in_intervals_filter(cls, day=None, lookup=None):
        u"""Возвращает фильтр для выборки записей по указанной дате.

        Условия фильтра включают записи, в интервал которых входит
        указанная дата.

        :params day: Дата, определяющая условия фильтра.
        :type day: datetime.date

        :param lookup: Список полей
        :type lookup: iterable

        :rtype: django.db.models.query_utils.Q
        """
        if day is None:
            day = datetime.date.today()

        from_name, to_name = cls.interval_field_names
        if lookup is not None:
            from_lookup = '__'.join(list(lookup) + [from_name])
            to_lookup = '__'.join(list(lookup) + [to_name])
        else:
            from_lookup, to_lookup = from_name, to_name

        from_filter = Q(**{from_lookup + '__lte': day})
        if cls._meta.get_field(from_name).null:
            # В модели допускается не указывать начало интервала
            from_filter |= Q(**{from_lookup + '__isnull': True})

        to_filter = Q(**{to_lookup + '__gte': day})
        if cls._meta.get_field(to_name).null:
            # В модели допускается не указывать конец интервала
            to_filter |= Q(**{to_lookup + '__isnull': True})

        result = from_filter & to_filter
        return result

    def interval_dates_error_message(self):
        u"""Возвращает сообщение о неправильных границах интервала.

        :rtype: unicode
        """
        date_from_name, date_to_name = self.interval_field_names
        result = u'Значение "{}" должно быть раньше, чем "{}"'.format(
            self._meta.get_field(date_from_name).verbose_name.capitalize(),
            self._meta.get_field(date_to_name).verbose_name.lower(),
        )
        return result

    def interval_intersected_error_message(self):
        u"""Возвращает сообщение о пересечении интервалов.

        :rtype: unicode
        """
        return u'Интервал пересекается с другими записями'

    @property
    def previous_interval(self):
        u"""Предыдущий интервал или None."""
        date_from_name, date_to_name = self.interval_field_names

        date_from = getattr(self, date_from_name)

        prev_intervals = self.__class__.objects.filter(
            **{'{}__lt'.format(date_to_name): date_from}
        ).order_by('-{}'.format(date_to_name))

        return prev_intervals[0] if prev_intervals else None

    @property
    def next_interval(self):
        u"""Следующий интервал или None."""
        date_from_name, date_to_name = self.interval_field_names

        date_to = getattr(self, date_to_name)

        next_intervals = self.__class__.objects.filter(
            **{'{}__gt'.format(date_from_name): date_to}
        ).order_by(date_from_name)

        return next_intervals[0] if next_intervals else None

    def clean(self):
        u"""Валидация данных об интервале."""
        errors = {}

        try:
            super(DateIntervalMixin, self).clean()
        except ValidationError as error:
            errors = error.update_error_dict(errors)

        date_from, date_to = self.interval_range

        # Дата начала интервала должна быть раньше даты окончания
        if (
            date_from is not None and
            date_to is not None and
            date_from > date_to
        ):
            errors.setdefault(NON_FIELD_ERRORS, []).append(
                self.interval_dates_error_message()
            )
        else:
            # Проверка пересения интервалов
            if self.get_intersected_query().exists():
                errors.setdefault(NON_FIELD_ERRORS, []).append(
                    self.interval_intersected_error_message()
                )

        if errors:
            raise ValidationError(errors)

    class Meta:
        abstract = True
